#############################################################################
#
#	core.py - Pyro Core Library
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################

import sys, time, string
import socket
import traceback
import Pyro.protocol
import util
from errors import *
from types import UnboundMethodType, MethodType, BuiltinMethodType, TupleType

Log=util.Log

#### Remote Invocation Flags (bit flags) ####

RIF_Varargs  = (1<<0)		# for '*args' syntax
RIF_Keywords = (1<<1)		# for '**keywords' syntax

#### Has init been performed already?
_init_server_done=0
_init_client_done=0


#############################################################################
#
#	ObjBase		- Server-side object implementation base class
#	              or master class with the actual object as delegate
#
#############################################################################

class ObjBase:
	def __init__(self):
		self.objectGUID=util.getBinaryGUID()
		self.delegate=None
		if Pyro.config.PYRO_MOBILE_CODE:
			self.codeValidator=lambda n,m,a: 1  # always accept
	def GUID(self):
		return self.objectGUID
	def setGUID(self, guid):			# used with persistent naming service
		self.objectGUID = guid
	def delegateTo(self,delegate):
		self.delegate=delegate
	def setDaemon(self, daemon):
		self.daemon=daemon
	def setCodeValidator(self, v):
		self.codeValidator=v
	def getDaemon(self):
		return self.daemon
	def getProxy(self):
		return self.daemon.getProxyForObj(self)
	def getAttrProxy(self):
		return self.daemon.getAttrProxyForObj(self)
	def Pyro_dyncall(self, method, flags, args):
		# find the method in this object, and call it with the supplied args.
		keywords={}
		if flags & RIF_Keywords:
			# reconstruct the varargs from a tuple like
			#  (a,b,(va1,va2,va3...),{kw1:?,...})
			keywords=args[-1]
			args=args[:-1]
		if flags & RIF_Varargs:
			# reconstruct the varargs from a tuple like (a,b,(va1,va2,va3...))
			args=args[:-1]+args[-1]
		# If the method is part of ObjBase, never call the delegate object because
		# that object doesn't implement that method. If you don't check this,
		# remote attributes won't work with delegates for instance, because the
		# delegate object doesn't implement _r_xa. (remote_xxxattr)
		if method in dir(ObjBase):  
			return apply(getattr(self,method),args,keywords)
		else:
			return apply(getattr(self.delegate or self,method),args,keywords)
	# remote getattr/setattr support:
	def _r_ha(self, attr):
		try:
			attr = getattr(self.delegate or self,attr)
			if type(attr) in (UnboundMethodType, MethodType, BuiltinMethodType):
				return 1 # method
		except:
			pass
		return 2 # attribute
	def _r_ga(self, attr):
		return getattr(self.delegate or self, attr)
	def _r_sa(self, attr, value):
		return setattr(self.delegate or self, attr, value)
	# remote code downloading support:
	def remote_supply_code(self, name, module, sourceaddr):
		if Pyro.config.PYRO_MOBILE_CODE and self.codeValidator(name,module,sourceaddr):
			import imp,marshal,new
			Log.msg('ObjBase','loading supplied code: ',name,'from',str(sourceaddr))
			env = new.module(name)
			if module[0:4]!=imp.get_magic():
				# compile source code
				code=compile(module,'<downloaded>','exec')
			else:
				# read bytecode
				code=marshal.loads(module[8:])
			exec code in env.__dict__
			sys.modules[name]=env
		else:
			Log.warn('ObjBase','attempt to supply code denied: ',name,'from',str(sourceaddr))
			raise PyroError('attempt to supply code denied')
			

#############################################################################
#
#	PyroURI		- Pyro Universal Resource Identifier
#
#	This class represents a Pyro URI (which consists of four parts,
#	a protocol identifier, an IP address, a portnumber, and an object ID.
#	
#	The URI can be converted to a string representation (str converter).
#	The URI can also be read back from such a string (reinitFromString).
#	The URI can be initialised from its parts (init).
#	The URI can be initialised from a string directly, if the init
#	 code detects a ':' and '/' in the host argument (which is then
#        assumed to be a string URI, not a host name/ IP address).
#
#############################################################################

class PyroURI:
	def __init__(self,host,objectID=0,port=0,protocol='PYRO'):
		# If the 'host' arg contains ':' and '/', assume it's a URI string.
		if string.find(host,':')>=0 and string.find(host,'/')>=0:
			self.reinitFromString(host)
		else:
			if Pyro.config.PYRO_DNS_URI:
				self.address = host
			else:
				self.address=Pyro.protocol.getIPAddress(host)
				if not self.address:
					raise URIError('unknown host')
			if port:
				self.port=port
			else:
				self.port=Pyro.config.PYRO_PORT
			self.protocol=protocol
			self.objectID=objectID
	def __str__(self):
		return self.protocol+'://'+self.address+':'+str(self.port)+  \
							'/'+util.bin2ascGUID(self.objectID)
	def __repr__(self):
		return '<PyroURI \''+str(self)+'\'>'
	def init(self,address,objectID,port=0,protocol='PYRO'):
		self.address=address
		self.objectID=objectID
		if port:
			self.port=port
		else:
			self.port=Pyro.config.PYRO_PORT
		self.protocol=protocol
	def reinitFromString(self,arg):
		arg=string.strip(arg)
		_argsave=arg
		colon = string.find(arg,':')
		try:
			if colon>=1:
				self.protocol = arg[:colon]
				if arg[colon+1:colon+3]=='//':
					arg=arg[colon+3:]
					colon = string.find(arg,':')
					if colon>=1:
						self.address=arg[:colon]
						arg=arg[colon+1:]
						slash=string.find(arg,'/')
						if slash>=1:
							self.port=int(arg[:slash])
							self.objectID=util.asc2binGUID(arg[slash+1:])
							return
					else:
						slash=string.find(arg,'/')
						if slash>=1:
							self.address=arg[:slash]
							self.port=Pyro.config.PYRO_PORT
							self.objectID=util.asc2binGUID(arg[slash+1:])
							return
		except ValueError, IndexError:
			pass
		Log.error('PyroURI','illegal URI format passed: '+_argsave)
		raise URIError('illegal URI format')


#############################################################################
#
#	DynamicProxy	- dynamic Pyro proxy
#
#	Can be used by clients to invoke objects for which they have no
#	precompiled proxy.
#
#############################################################################

def getProxyForURI(URI):
	return DynamicProxy(URI)
def getAttrProxyForURI(URI):
	return DynamicProxyWithAttrs(URI)

class DynamicProxy:
	def __init__(self, URI):
		self.URI = URI
		self.objectID = URI.objectID
		self.adapter = None				# delay adapter binding
		self._name=[]
	def __copy__(self):
		c=DynamicProxy(self.URI)
		c.adapter=None
		return c
	def __getattr__(self, name):
		# allows it to be safely pickeled
		if name not in ("__getinitargs__", "__eq__"):
			self._name.append(name)
			return self.__invokePYRO__
		raise AttributeError()
	def __repr__(self):
		return "<"+self.__class__.__name__+" instance at "+hex(id(self))+">"
	def __str__(self):
		return repr(self)
	def __hash__(self):
		# makes it possible to use this class as a key in a dict
		return hash(self.objectID)

	def __invokePYRO__(self, *vargs, **kargs):
		if self.adapter is None:
			# delayed adapter binding.
			self.adapter = Pyro.protocol.getProtocolAdapter(self.URI.protocol)
			self.adapter.bindToURI(self.URI)
		name = self._name[-1]
		del self._name[-1]
		return self.adapter.remoteInvocation(name,
						RIF_Varargs|RIF_Keywords, vargs, kargs)
	# pickling support:
	def __getstate__(self):
		temp = {}
		for key in self.__dict__.keys():
			if key != "adapter":
				temp[key] = self.__dict__[key]
		return temp
	def __setstate__(self, value):
		for key in value.keys():
			self.__dict__[key] = value[key]
		self.adapter = None

	
class DynamicProxyWithAttrs(DynamicProxy):
	def __init__(self, URI):
		# first set the list of 'local' attrs for __setattr__
		self.__dict__["_local_attrs"] = ("_local_attrs","URI", "objectID", "adapter", "_name", "attr_cache")
		DynamicProxy.__init__(self, URI)
		self.attr_cache = {}
	def _r_ga(self, attr, value=0):
		if value: self._name.append("_r_ga")
		else: self._name.append("_r_ha")
		return self.__invokePYRO__(attr)
	def findattr(self, attr):
		if self.attr_cache.has_key(attr):
			return self.attr_cache[attr]
		# look it up and cache the value
		self.attr_cache[attr] = self._r_ga(attr)
		return self.attr_cache[attr]
	def __setattr__(self, attr, value):
		if attr in self.__dict__["_local_attrs"]:
			self.__dict__[attr]=value
		else:
			result = self.findattr(attr)
			if result==2: # attribute
				self._name.append("_r_sa")
				return self.__invokePYRO__(attr,value)
			else:
				raise AttributeError('not an attribute')
	def __getattr__(self, attr):
		# allows it to be safely pickeled
		if attr != "__getinitargs__":
			result=self.findattr(attr)
			if result==1: # method
				self._name.append(attr)
				return self.__invokePYRO__
			elif result:
				return self._r_ga(attr, 1)
		raise AttributeError()
		

#############################################################################
#
#	Daemon		- server-side Pyro daemon
#
#	Accepts and dispatches incoming Pyro method calls.
#
#############################################################################

class Daemon(Pyro.protocol.TCPServer):
	def __init__(self,protocol='PYRO',host='',port=0,norange=0,publishhost=None):
		self.implementations={}
		if port:
			self.port = port
		else:
			self.port = Pyro.config.PYRO_PORT
		if norange:
			portrange=1
		else:
			portrange=Pyro.config.PYRO_PORT_RANGE
		if not publishhost:
			publishhost=host
		errormsg=''
		for i in range(portrange):
			try:
				Pyro.protocol.TCPServer.__init__(self, self.port, host, Pyro.config.PYRO_MULTITHREADED)
				self.hostname = publishhost or Pyro.protocol.getHostname()	
				self.NameServer = None
				self.protocol = protocol
				self.adapter = Pyro.protocol.getProtocolAdapter(protocol)
				self.adapter.setDaemon(self)
				return
			except socket.error,msg:
				errormsg=msg
				self.port = self.port+1
		Log.error('Daemon','Couldn\'t start Pyro daemon: ' +str(errormsg))
		raise DaemonError('Couldn\'t start Pyro daemon: ' +str(errormsg))
	
	def __del__(self):
		# server shutting down, unregister all known objects in the NS
		for o in self.implementations.keys():
			self.NameServer.unregister(self.implementations[o][1])

	def __str__(self):
		return '<Pyro Daemon on '+self.hostname+':'+str(self.port)+'>'

	def useNameServer(self,NS):
		self.NameServer=NS

	def getProxyForObj(self, obj):
		return DynamicProxy( PyroURI(self.hostname, obj.GUID(),
					  protocol=self.protocol, port=self.port) )
	def getAttrProxyForObj(self, obj):
		return DynamicProxyWithAttrs ( PyroURI(self.hostname, obj.GUID(),
					  protocol=self.protocol, port=self.port) )

	def connectPersistent(self, object, name=None):
		# when a persistent entry is found in the NS, that URI is
		# used instead of the supplied one, if the address matches.
		if name and self.NameServer:
			try:
				newURI = PyroURI(self.hostname, object.GUID(),
					  protocol=self.protocol, port=self.port)
				URI=self.NameServer.resolve(name)
				if (URI.protocol,URI.address,URI.port)==(newURI.protocol,newURI.address,newURI.port):
					object.setGUID(URI.objectID)
					# enter the (object,name) in the known impl. dictionary
					self.implementations[object.GUID()]=(object,name)
					object.setDaemon(self)
					return URI
				else:
					# name exists, but address etc. is wrong. Remove it.
					# then continue so it wil be re-registered.
					try: self.NameServer.unregister(name)
					except NamingError: pass
			except NamingError:
				pass
		return self.connect(object, name)

	def connect(self, object, name=None):
		URI = PyroURI(self.hostname, object.GUID(),
					  protocol=self.protocol, port=self.port)
		# if not transient, register the object with the NS
		if name:
			if self.NameServer:
				self.NameServer.register(name, URI)
			else:
				Log.warn('Daemon','connecting object without naming service specified:',name)
		# enter the (object,name) in the known implementations dictionary
		self.implementations[object.GUID()]=(object,name)
		object.setDaemon(self)
		return URI

	def disconnect(self,object):
		try:
			if self.implementations[object.GUID()][1]:
				# only unregister with NS if it had a name (was not transient)
				self.NameServer.unregister(self.implementations[object.GUID()][1])
			del self.implementations[object.GUID()]
		finally:
			object.setDaemon(None)

	def handleError(self,conn):			# overridden from TCPServer
		(exc_type, exc_value, exc_trb) = sys.exc_info()
		if exc_type in(ProtocolError, socket.error):
			# ProtocolError or socket.error, shut down the connection
			# XXX is this wat we want???
			Log.error('Daemon','Due to network error: shutting down connection with',conn.addr)
			self.removeConnection(conn)
		else:
			# This code used to distinguish between 'accidents' and
			# 'on purpose' exceptions. No longer. Every exception
			# is treated the same.
			tb = traceback.format_exception(exc_type, exc_value, exc_trb)[-5:]
			out=''
			for line in tb:
				out=out+line
			Log.warn('Daemon', 'Exception during processing of request from',
				conn.addr,' type',exc_type,
				'\n--- partial traceback of this exception follows:\n',
				out,'\n--- end of traceback')
			self.adapter.returnException(conn.sock,exc_value)
	
	def handleRequest(self,conn):
		try:
			self.adapter.handleInvocation(conn)
		except SocketClosedError:
			# other side has closed connection
			self.removeConnection(conn)


#############################################################################
#
#	Client/Server Init code
#
#############################################################################

import time, string

def _initGeneric_pre():
	if Pyro.config.PYRO_TRACELEVEL == 0: return
	try:
		out='\n'+'-'*60+' NEW SESSION\n'+time.ctime(time.time())+ \
			'   Pyro Initializing, version '+Pyro.PYRO_VERSION+'\n'
		Log.raw(out)
	except IOError,e:
		sys.stderr.write('PYRO: Can\'t write the tracefile '+Pyro.config.PYRO_LOGFILE.name+'\n'+e)

def _initGeneric_post():
	if Pyro.config.PYRO_TRACELEVEL == 0: return
	try:
		out='Configuration settings are as follows:\n'
		for item in dir(Pyro.config):
			if item[0:4]=='PYRO':
				out=out+item+' = '+str(Pyro.config.__dict__[item])+'\n'
		out=out+'Init done.\n'+'-'*70+'\n'
		Log.raw(out)
	except IOError,e:
		pass

def initClient(banner=1):
	global _init_client_done
	if _init_client_done: return
	_initGeneric_pre()
	if Pyro.config.PYRO_TRACELEVEL >0: Log.raw('This is initClient.\n')
	_initGeneric_post()
	if banner:
		print 'Pyro Client Initialized. Using Pyro V'+Pyro.PYRO_VERSION
	_init_client_done=1
	
def initServer(banner=1):
	global _init_server_done
	if _init_server_done: return
	_initGeneric_pre()
	if Pyro.config.PYRO_TRACELEVEL >0: Log.raw('This is initServer.\n')
	_initGeneric_post()
	if banner:
		print 'Pyro Server Initialized. Using Pyro V'+Pyro.PYRO_VERSION
	_init_server_done=1

