#############################################################################
#
#	protocol.py - Pyro Protocol Adapters
#
#	This is part of "Pyro" - Python Remote Objects
#	which is (c) Irmen de Jong - irmen@users.sourceforge.net
#
#############################################################################

import select, socket, struct, os, time, sys
import Pyro.util
from errors import *
from errors import _InternalNoModuleError
pickle = Pyro.util.pickle
Log = Pyro.util.Log

if Pyro.util.supports_multithreading():
	from threading import Thread
	_has_threading = 1
else:
	_has_threading = 0

if Pyro.util.supports_compression():
	import zlib
	_has_compression = 1
else:
	_has_compression = 0

#------ Get the hostname (possibly of other machines) (returns None on error)
def getHostname(ip=None):
	try:
		if ip:
			(hn,alias,ips) = socket.gethostbyaddr(ip)
			return hn
		else:
			return socket.gethostname()
	except socket.error:
		return None

#------ Get IP address (return None on error)
def getIPAddress(host=None):
	try:
		return socket.gethostbyname(host or getHostname())
	except socket.error:
		return None


#------ Socket helper functions for sending and receiving data correctly.

# Receive a precise number of bytes from a socket. Raises the
# SocketClosedError if  that number of bytes was not available.
# (the connection has probably been closed then).
# Never will this function return an empty message (if size>0).
# We need this because 'recv' isn't guaranteed to return all desired
# bytes in one call, for instance, when network load is high.
def sock_recvmsg(sock, size):
	msg=''
	try:
		while len(msg)<size:
			chunk=sock.recv(size-len(msg))
			if not chunk:
				raise SocketClosedError('connection lost')
			msg=msg+chunk
		return msg
	except socket.error,x:
		raise SocketClosedError('connection lost')

# Send a message over a socket. Raises SocketClosedError if the msg
# couldn't be sent (the connection has probably been lost then).
# We need this because 'send' isn't guaranteed to send all desired
# bytes in one call, for instance, when network load is high.
def sock_sendmsg(sock, msg):
	size=len(msg)
	total=0
	try:
		while total<size:
			sent=sock.send(msg[total:])
			if sent==0:
				raise SocketClosedError('connection lost')
			total=total+sent
	except socket.error,x:
		raise SocketClosedError('connection lost')

# set socket option to try to re-use a server port if possible
def set_reuse_addr(sock):
	try:
		sock.setsockopt ( socket.SOL_SOCKET, socket.SO_REUSEADDR,
			self.socket.getsockopt (socket.SOL_SOCKET, socket.SO_REUSEADDR) | 1)
	except:
		pass


#------ PYRO: adapter (default Pyro wire protocol)
#------ This adapter is for protocol version 2.0 ONLY
# Future adapters could be downwards compatible and more flexible.
# Notice that when the header and body format changes, we have
# to increment the protocol VERSION. A revision change is only when
# something inside the body is different but the message as a whole
# can still be received.

PFLG_COMPRESSED = 0x01		# protocol flag: compressed body

class PYROAdapter:
	deniedReasons={ 0:'unspecified reason',1:'server too busy', 2:'host blocked', 3:'security reasons'}
	def __init__(self):
		self.daemon = None			# the Pyro daemon which uses us
		self.headerFmt = '!4sBBHlB'	# version 2 header
		self.headerID = 'PYRO'
		self.acceptMSG= 'GRANTED'	# must be same length as denyMSG
		self.denyMSG=   'DENIED'	# must be same length as acceptMSG
		                            # (note that a 1-character code is appended)
		self.headerSize = struct.calcsize(self.headerFmt)
		self.version, self.revision = (2,0)	# version 2.0
	def setDaemon(self, daemon):
		self.daemon = daemon
		Log.msg('PYROAdapter','adapter daemon set to',str(daemon))
	def sendAccept(self, sock):		# called by TCPServer
		sock_sendmsg(sock, self.acceptMSG)
	def sendDeny(self, sock, reasonCode=0):		# called by TCPServer
		sock_sendmsg(sock, self.denyMSG+str(reasonCode)[0])
	def bindToURI(self,URI):
		if URI.protocol!='PYRO':
			Log.error('PYROAdapter','incompatible protocol in URI:',URI.protocol)
			raise ProtocolError('incompatible protocol in URI')
		try:
			self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
			self.sock.connect((URI.address, URI.port))
			msg = sock_recvmsg(self.sock, len(self.acceptMSG))
			if msg==self.acceptMSG:
				self.URI=URI
				Log.msg('PYROAdapter','connected to',str(URI))
			elif msg[:len(self.denyMSG)]==self.denyMSG:
				try:
					raise ConnectionDeniedError(self.deniedReasons[int(msg[-1])])
				except (KeyError,ValueError):
					raise ConnectionDeniedError('invalid response')
		except socket.error:
			Log.msg('PYROAdapter','connection failed to URI',str(URI))
			raise ProtocolError('connection failed')

	def rebindURI(self, tries=sys.maxint, wait=1):
		t=0
		while t<tries:
			try:
				self.bindToURI(self.URI)
				return
			except ProtocolError:
				t=t+1
				if t<tries:
					time.sleep(wait)
		raise SocketClosedError('connection lost')
		
	def createMsg(self, body):
		pflgs=0
		if _has_compression and Pyro.config.PYRO_COMPRESSION:
			before=len(body)
			bz=zlib.compress(body)
			if len(bz)<before:
				pflgs=PFLG_COMPRESSED
				body=bz
		return struct.pack(self.headerFmt, self.headerID,
					self.version, self.revision,
					self.headerSize, len(body), pflgs) + body

	def remoteInvocation(self, method, flags, *args):
		body=pickle.dumps((self.URI.objectID,method,flags,args),Pyro.config.PYRO_BINARY_PICKLE)
		sock_sendmsg(self.sock, self.createMsg(body))
		ver,rev,answer = self.receiveMsg(self.sock)
		answer = pickle.loads(answer)
		if isinstance(answer,PyroExceptionCapsule):
			if isinstance(answer.excObj,_InternalNoModuleError):
				# server couldn't load module, supply it
				m=__import__(answer.excObj.modulename)
				m=open(m.__file__,'rb').read()
				self.remoteInvocation('remote_supply_code',0,answer.excObj.modulename, m, self.sock.getsockname())
				# retry the method invocation
				return apply(self.remoteInvocation,(method, flags)+args)
			else:
				# we have an encapsulated exception, raise it again.
				answer.raiseEx()
		return answer

	def receiveMsg(self,sock):
		msg=sock_recvmsg(sock, self.headerSize)
		(id, ver, rev, hsiz, bsiz, pflags) = struct.unpack(self.headerFmt,msg)
		if ver>self.version:
			Log.error('PYROAdapter','incompatible protocol version')
			raise ProtocolError('incompatible protocol version')
		if id!=self.headerID or hsiz!=self.headerSize:
			print 'INVALID HEADER: ',sock.getsockname(),(id, ver, rev, hsiz, bsiz,pflags) # XXX
			Log.error('PYROAdapter','invalid header')
			raise ProtocolError('invalid header')
		body=sock_recvmsg(sock, bsiz)
		if pflags&PFLG_COMPRESSED:
			if _has_compression:
				body=zlib.decompress(body)
			else:
				raise ProtocolError('compression not supported')
		return ver,rev,body

	def handleInvocation(self,conn):
		ver,rev,body = self.receiveMsg(conn.sock)
		if ver!=self.version or rev!=self.revision:
			Log.error('PYROAdapter','incompatible protocol version')
			self.returnException(conn.sock, ProtocolError('incompatible protocol version'))
			conn.sock.shutdown(2) # no further send/receives
			return
		# Unpickle the request, which is a tuple:
		#  (object ID, method name, flags, (arg1,arg2,...))
		try:
			req=pickle.loads(body)
		except ImportError,x:
			if Pyro.config.PYRO_MOBILE_CODE:
				# return a special exception that will be processed by client;
				# it will call the internal 'remote_supply_code' member
				self.returnException(conn.sock, _InternalNoModuleError(x.args[0][16:]))
			else:
				Log.error('PYROAdapter','code problem with incoming object: '+str(x))
				self.returnException(conn.sock, apply(NoModuleError,x.args))
			return
		try:
			# find the object in the implementation database of our daemon
			o = self.daemon.implementations[req[0]]
		except KeyError,x:
			Log.warn('PYROAdapter','Invocation to unknown object ignored:',x)
			self.returnException(conn.sock, ProtocolError('unknown object ID'))
			conn.sock.shutdown(2) # no further send/receives
			return
		else:
			# call the method on this object (dispatch in a thread if desired)
			if self.daemon.threaded:
				Thread(target=self._handleInvoc_thread, args=(o[0],req,conn)).start()
			else:
				self._handleInvoc_thread(o[0],req,conn)
	def _handleInvoc_thread(self, proxy, req, conn):
		try:
			res = proxy.Pyro_dyncall(req[1],req[2],req[3])	# (method,flags,args)
			# reply the result to the caller
			body=pickle.dumps(res,Pyro.config.PYRO_BINARY_PICKLE)
			sock_sendmsg(conn.sock, self.createMsg(body))
		except:
			self.daemon.handleError(conn)

	def returnException(self, sock, exc):
		# return an encapsulated exception to the client
		try:
			body=pickle.dumps(PyroExceptionCapsule(exc),Pyro.config.PYRO_BINARY_PICKLE)
		except Exception,x:
			# hmm, pickling the exception failed... pickle the string instead
			body=pickle.dumps(PyroExceptionCapsule(PyroError(str(x))),Pyro.config.PYRO_BINARY_PICKLE)
		sock_sendmsg(sock, self.createMsg(body))
			
	
		
def getProtocolAdapter(protocol):
	if protocol=='PYRO':
		return PYROAdapter()
	else:
		Log.error('getProtocolAdapter','unsupported protocol:',protocol)
		raise ProtocolError('unsupported protocol')


#-------- TCPConnection object for TCPServer class
class TCPConnection:
	def __init__(self, sock, addr):
		self.sock = sock
		self.addr = addr
	def fileno(self):
		return self.sock.fileno()
	def close(self):
		self.sock.close()

#-------- new connection validators:
#-------- limitingConnValidator checks max number of connections
class limitingConnValidator:
	def __init__(self, tcpserver):
		self.tcpserver = tcpserver
	def accept(self,connection):
		if len(self.tcpserver.connections)>=Pyro.config.PYRO_MAXCONNECTIONS:
			Log.msg('limitingConnValidator','Too many open connections, closing',connection.addr,'#conns=',len(self.tcpserver.connections))
			return (0, 1) # reason 1 = server full
		return (1,0)

#-------- TCPServer base class
class TCPServer:
	def __init__(self, port, host='', threaded=_has_threading):
		self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		set_reuse_addr(self.sock)
		self.sock.bind((host,port))
		self.sock.listen(5)
		self.connections = []
		self.setParamsForLoop(5)
		self.threaded = threaded
		self.newConnValidator=limitingConnValidator(self)
	def __del__(self):
		if len(self.connections)>0:
			Log.warn('TCPServer','Shutting down but there are still',len(self.connections),'active connections')
			
	def setParamsForLoop(self, timeout, others=[], callback=None):
		self.timeout=timeout
		self.others=others
		self.callback=callback
	def connectionLoop(self, condition=lambda:1):
		while condition():
			self.handleRequests(self.timeout,self.others,self.callback)
	def setNewConnectionValidator(self,validator):
		self.newConnValidator=validator
	def newConnection(self,sock):
		csock, addr = sock.accept()
		conn=TCPConnection(csock,addr)
		(ok,reasonCode) = self.newConnValidator.accept(conn)
		if ok:
			self.adapter.sendAccept(csock)
			self.connections.append(conn)
			Log.msg('TCPServer','new connection from',addr,'#conns=',len(self.connections))
		else:
			# log entry has already been written by newConnValidator
			self.adapter.sendDeny(csock,reasonCode)
			conn.close()

	def handleRequests(self, timeout=None, others=[], callback=None):
		handledRequest=0	
		activeCnt=1
		while activeCnt and not handledRequest:
			socklist = self.connections+[self.sock]+others
			if timeout==None:
				ins,outs,exs = select.select(socklist,[],[])
			else:
				ins,outs,exs = select.select(socklist,[],[],timeout)
			activeCnt=len(ins)
			if self.sock in ins:
				self.newConnection(self.sock)
				ins.remove(self.sock)
			for c in ins[0:]:
				if isinstance(c,TCPConnection):
					try:
						self.handleRequest(c)
					except:
						self.handleError(c)
					ins.remove(c)
					handledRequest=1
			if ins and callback:
				# the 'others' must have fired...
				callback(ins)

	def handleError(self,conn):
		# Default implementation. Most likely you have to override
		# this in your subclass.
		print '-'*40
		print 'Exception happened during processing of request from',conn.addr
		import traceback
		traceback.print_exc()
		print '-'*40

	# to be called if a dropped connection is detected:
	def removeConnection(self, conn):
		if conn in self.connections:
			self.connections.remove(conn)
			Log.msg('TCPServer','removed connection with',conn.addr,' #conns=',len(self.connections))

